import random
from enum import Enum

verbose = False
saturation = 70 # global saturation of output colors

class Outcome(Enum):
	FAIL = 6	  #  6-
	PARTIAL = 9  #  7-9
	SUCCESS = 11 # 10-11
	TRIUMPH = 99 # 12+

	def __lt__(self, other):
		if self.__class__ is other.__class__:
			return self.value < other.value
		return NotImplemented

	def __ge__(self, other):
		if self.__class__ is other.__class__:
			return self.value >= other.value
		return NotImplemented

	def next(self):
		for outcome in Outcome:
			if (self < outcome):
				return outcome
		return self

outcomeLookup = {
	'combined' : (Outcome.FAIL,Outcome.PARTIAL,Outcome.SUCCESS),
	'hue' : {
		Outcome.FAIL: 0,
		Outcome.PARTIAL: 60,
		Outcome.SUCCESS: 120,
		Outcome.TRIUMPH: 180
	},
	'label': {
		Outcome.FAIL: '6&minus;',
		Outcome.PARTIAL: '7&minus;9',
		Outcome.SUCCESS: '10+',
		Outcome.TRIUMPH: '12+'
	},
	'xlabel': {
		Outcome.FAIL: '6&minus;',
		Outcome.PARTIAL: '7&minus;9',
		Outcome.SUCCESS: '10&minus;11',
		Outcome.TRIUMPH: '12+'
	}
}

class Results:
	def __init__(self):
		self.outcomes = {}
		for aidOutcome in Outcome:
			self.outcomes['total'] = {}
			self.outcomes[aidOutcome] = {}
			self.outcomes[aidOutcome]['total'] = 0
			for testOutcome in Outcome:
				self.outcomes['total'][testOutcome] = 0
				self.outcomes[aidOutcome][testOutcome] = 0

	def trackResult(self, aidOutcome, testOutcome):
		self.outcomes[aidOutcome][testOutcome] += 1
		self.outcomes['total'][testOutcome] += 1
		self.outcomes[aidOutcome]['total'] += 1

	def outputCell(self,total,combineTriumph,aidOutcome,testOutcome):
		resultsInCell = self.getResult(aidOutcome,testOutcome,combineTriumph)
		text = self.cellFromResult(aidOutcome,testOutcome,resultsInCell,total)
		return text

	def getResult(self,aidOutcome,testOutcome,combineTriumph):
		resultsInCell = self.outcomes[aidOutcome][testOutcome]
		if (combineTriumph):
			# Combine the triumph results into the success cells
			if aidOutcome == Outcome.SUCCESS:
				resultsInCell += self.outcomes[Outcome.TRIUMPH][testOutcome]
			if testOutcome == Outcome.SUCCESS:
				resultsInCell += self.outcomes[aidOutcome][Outcome.TRIUMPH]
				if aidOutcome == Outcome.SUCCESS:
					resultsInCell += self.outcomes[Outcome.TRIUMPH][Outcome.TRIUMPH]
		return resultsInCell

	def cellFromResult(self,aidOutcome,testOutcome,resultsInCell,total):
		cellPercent = resultsInCell * 100 / total
		lightness = 100 - (cellPercent * 1.3) # Make the color slightly more dramatic
		hue = (outcomeLookup['hue'][aidOutcome] + outcomeLookup['hue'][testOutcome]) / 2
		return '<td style="background-color: hsl(%d,%d%%,%d%%)">%.1f</td>' % (hue,saturation,lightness,round(cellPercent,1))

class Counter:
	def __init__(self,aidBonus,testBonus):
		self.aidBonus = aidBonus
		self.testBonus = testBonus
		self.noInteraction = Results()
		self.resultsAsWritten = Results()
		self.resultsVariant = Results()

	def roll(self):
		aidRoll = self.getRoll(self.aidBonus)
		testRoll = self.getRoll(self.testBonus)
		self.applyNoInteraction(aidRoll,testRoll)
		self.applyAsWritten(aidRoll,testRoll)
		self.applyVariant(aidRoll,testRoll)

	def getRoll(self, modifier):
		return random.randint(1, 6) + random.randint(1, 6) + modifier

	def getOutcome(self, roll):
		for outcome in Outcome:
			if roll <= outcome.value:
				return outcome
		return Outcome.TRIUMPH

	def applyNoInteraction(self,aidRoll,testRoll):
		aidOutcome = self.getOutcome(aidRoll)
		testOutcome = self.getOutcome(testRoll)
		self.noInteraction.trackResult(aidOutcome, testOutcome)

	def applyAsWritten(self,aidRoll,testRoll):
		aidOutcome = self.getOutcome(aidRoll)
		if aidOutcome >= Outcome.PARTIAL:
			testRoll = testRoll + 1
		testOutcome = self.getOutcome(testRoll)
		self.resultsAsWritten.trackResult(aidOutcome, testOutcome)

	def applyVariant(self,aidRoll,testRoll):
		aidOutcome = self.getOutcome(aidRoll)
		if aidOutcome == Outcome.PARTIAL:
			testRoll = testRoll + 1
		testOutcome = self.getOutcome(testRoll)
		if aidOutcome > Outcome.PARTIAL:
			testOutcome = testOutcome.next()
		self.resultsVariant.trackResult(aidOutcome, testOutcome)

	def outputNoInteractionCell(self,total,combineTriumph,aidOutcome,testOutcome):
		return self.noInteraction.outputCell(total,combineTriumph,aidOutcome,testOutcome)

	def outputAsWrittenCell(self,total,combineTriumph,aidOutcome,testOutcome):
		return self.resultsAsWritten.outputCell(total,combineTriumph,aidOutcome,testOutcome)

	def outputVariantCell(self,total,combineTriumph,aidOutcome,testOutcome):
		return self.resultsVariant.outputCell(total,combineTriumph,aidOutcome,testOutcome)

	def outputAidEffectCell(self,total,combineTriumph,aidOutcome,testOutcome):
		return self.outputDifference(self.noInteraction,self.resultsAsWritten,total,combineTriumph,aidOutcome,testOutcome)

	def outputDifferenceCell(self,total,combineTriumph,aidOutcome,testOutcome):
		return self.outputDifference(self.resultsAsWritten,self.resultsVariant,total,combineTriumph,aidOutcome,testOutcome)

	def outputDifference(self,results1,results2,total,combineTriumph,aidOutcome,testOutcome):
		result1 = results1.getResult(aidOutcome,testOutcome,combineTriumph)
		result2 = results2.getResult(aidOutcome,testOutcome,combineTriumph)
		resultsInCell = result2 - result1
		cellPercent = resultsInCell * 100 / total
		magnitude = abs(cellPercent)
		text = ''
		if magnitude > 0.05:
			lightness = (100 - (5 * magnitude))
			hue = (outcomeLookup['hue'][aidOutcome] + outcomeLookup['hue'][testOutcome]) / 2
			text += '<td style="background-color: hsl(%d,%d%%,%d%%)">%+.1f</td>' % (hue,saturation,lightness,round(cellPercent,1))
		else:
			text += '<td>&nbsp;</td>'
		return text

	def output(self,total):
		print ('Aid bonus +%d  Test bonus +%d' % (self.aidBonus,self.testBonus))
		print ('As Written')
		print ('Variation')


def writeTable(combineTriumph,cellRenderingMethodName):
	useOutcomes = Outcome
	labelKey = 'xlabel'
	columnLabel = 'Test Bonus'
	rowLabel = 'Aid<br/>Bonus'
	if cellRenderingMethodName == 'outputNoInteractionCell':
		columnLabel = 'Second 2d6 Roll'
		rowLabel = 'First<br/>2d6<br/>Roll'
	if combineTriumph:
		useOutcomes = outcomeLookup['combined']
		labelKey = 'label'
	testSpan = len(Counts[0]) * len(useOutcomes)
	aidSpan = len(Counts) * len(useOutcomes)

	print ('<table class="results %s">' % (cellRenderingMethodName))
	print ('<tr><td rowspan="3" colspan="3"></td><th class="roll2" colspan="%d">%s</th></tr>' % (testSpan,columnLabel))
	bonusLine = '<tr>'
	resultLine = '<tr>'
	for testBonus in testRange:
		bonusLine += '<th class="roll2" colspan="%d">%+d</th>' % (len(useOutcomes),testBonus)
		for testOutcome in useOutcomes:
			resultLine += '<th>%s</th>' % (outcomeLookup[labelKey][testOutcome])
	bonusLine += '</tr>'
	print (bonusLine)
	resultLine += '</tr>'
	print (resultLine)
	for aidBonus in aidRange:
		for aidOutcome in useOutcomes:
			line = '<tr>'
			if (aidOutcome == Outcome.FAIL):
				if (aidBonus == 0):
					line += '<th class="roll1" rowspan="%d">%s</td>' % (aidSpan,rowLabel)
				line += '<th class="roll1" rowspan="%d">+%d</th>' % (len(useOutcomes),aidBonus)
			line += '<th>%s</th>' % (outcomeLookup[labelKey][aidOutcome])
			for testBonus in testRange:
				counter = Counts[aidBonus][testBonus]
				cellRenderMethod = getattr(counter, cellRenderingMethodName)
				for testOutcome in useOutcomes:
					line += cellRenderMethod(repeat,combineTriumph,aidOutcome,testOutcome)
			line += '</tr>'
			print (line)
	print ('</table>\n')

def writeBaseTable(combineTriumph):
	result = Counts[0][0].noInteraction
	print ('<table class="results tworolls">')
	outcomes = Outcome
	labelKey = 'xlabel'
	if combineTriumph:
		outcomes = outcomeLookup['combined']
		labelKey = 'label'
	print ('<tr><td rowspan="2" colspan="2"></td><th class="roll2" colspan="%d">Second 2d6 Roll</th></tr>' % (len(outcomes)))
	print ('<tr>')
	for testOutcome in outcomes:
		print ('<th>%s</th>' % (outcomeLookup[labelKey][testOutcome]))
	print ('</tr>')
	for aidOutcome in outcomes:
		print ('<tr>')
		if aidOutcome == Outcome.FAIL:
			print('<th class="roll1" rowspan="%d">First<br/>2d6<br/>Roll</th>' % (len(outcomes)))
		print('<th>%s</th>' % (outcomeLookup[labelKey][aidOutcome]))
		for testOutcome in outcomes:
			print(result.outputCell(repeat,combineTriumph,aidOutcome,testOutcome))
		print ('</tr>')
	print ('</table>')
	
def writeSummaryTable():
	combineTriumph = False
	useOutcomes = Outcome
	labelKey = 'xlabel'
	columnLabel = 'Test Bonus'
	testSpan = len(Counts[0]) * len(useOutcomes)
	aidSpan = len(Counts)

	print ('<table class="results testSummary">')
	print ('<tr><td rowspan="3" colspan="2"></td><th class="roll2" colspan="%d">%s</th></tr>' % (testSpan,columnLabel))
	bonusLine = '<tr>'
	resultLine = '<tr>'
	for testBonus in testRange:
		bonusLine += '<th class="roll2" colspan="%d">%+d</th>' % (len(useOutcomes),testBonus)
		for testOutcome in useOutcomes:
			resultLine += '<th>%s</th>' % (outcomeLookup[labelKey][testOutcome])
	bonusLine += '</tr>'
	print (bonusLine)
	resultLine += '</tr>'
	print (resultLine)
	
	line = '<tr><th class="roll1" colspan="2">Without aid</td>'
	for testBonus in testRange:
		for testOutcome in useOutcomes:
			counter = Counts[0][testBonus]
			resultsInCell = counter.noInteraction.getResult('total',testOutcome,combineTriumph)
			cellPercent = resultsInCell * 100 / repeat
			lightness = 100 - (cellPercent * 0.6) # tone down the color a little
			hue = outcomeLookup['hue'][testOutcome]
			line += '<td style="background-color: hsl(%d,%d%%,%d%%)">%.1f</td>' % (hue,saturation,lightness,round(cellPercent,1))
	line += '</tr>'
	print (line)
	
	rowLabel = 'Standard Aid'
	for resultsAttr in ('resultsAsWritten', 'resultsVariant'):
		for aidBonus in aidRange:
			line = '<tr>'
			if (aidBonus == 0):
				line += '<th class="roll1" rowspan="%d">%s</td>' % (aidSpan,rowLabel)
			line += '<th class="roll1">+%d</th>' % (aidBonus)
			for testBonus in testRange:
				counter = Counts[aidBonus][testBonus]
				results = getattr(counter, resultsAttr)
				for testOutcome in useOutcomes:
					resultsInCell = results.getResult('total',testOutcome,combineTriumph)
					cellPercent = resultsInCell * 100 / repeat
					lightness = 100 - (cellPercent * 0.6) # tone down the color a little
					hue = outcomeLookup['hue'][testOutcome]
					line += '<td style="background-color: hsl(%d,%d%%,%d%%)">%.1f</td>' % (hue,saturation,lightness,round(cellPercent,1))
			line += '</tr>'
			print (line)
		rowLabel = '<cite>Fourth World</cite> Aid'
	print ('</table>\n')

Counts = {}
repeat = 100 # quick testing
#repeat = 100000 # longer running but more accurate
#repeat = 1000000 # overkill
random.seed()
aidRange = range(0,4)
testRange = range(-1,4)
for aidBonus in aidRange:
	Counts[aidBonus] = {}
	for testBonus in testRange:
		counter = Counter(aidBonus,testBonus)
		Counts[aidBonus][testBonus] = counter
		for count in range(0, repeat):
			counter.roll()

writeBaseTable(True)
writeBaseTable(False)
writeTable(True, 'outputNoInteractionCell')
writeTable(True, 'outputAsWrittenCell')
writeTable(False, 'outputAsWrittenCell')
writeTable(False, 'outputAidEffectCell')
writeTable(False, 'outputVariantCell')
writeTable(False, 'outputDifferenceCell')
writeSummaryTable()